import pathlib

import numpy as np
from cpymad.madx import Madx

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_ps_multiturn_twiss(test_context):

    test_data_folder = pathlib.Path(
            __file__).parent.joinpath('../test_data').absolute()

    mad = Madx(stdout=False)
    mad.input("""
    beam, particle=proton, pc = 14.0;
    BRHO      = BEAM->PC * 3.3356;
    """)
    mad.call(str(test_data_folder / "ps_sftpro/ps.seq"))
    mad.call(str(test_data_folder / "ps_sftpro/ps_hs_sftpro.str"))
    mad.use('ps')

    line = xt.Line.from_madx_sequence(mad.sequence.ps, allow_thick=True,
                                    deferred_expressions=True,
                                    )
    line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV,
                                        q0=1, gamma0=mad.sequence.ps.beam.gamma)
    line.twiss_default['method'] = '4d'
    line.build_tracker(test_context)

    tw = line.twiss()

    opt = line.match(
        solve=False,
        vary=[xt.VaryList(['kf', 'kd'], step=1e-5)],
        targets=[xt.TargetSet(qx=6.255278, qy=6.29826, tol=1e-7)],
    )
    opt.solve()


    r0 = np.linspace(0, 100, 50)
    p = line.build_particles(
        x_norm=r0*np.cos(np.pi/20.),
        px_norm=r0*np.sin(np.pi/20.),
        nemitt_x=1e-6, nemitt_y=1e-6)

    line.track(p, num_turns=1000, turn_by_turn_monitor=True)
    mon = line.record_last_track

    tw_mt = line.twiss(co_guess={'x': 0.025}, num_turns=4)
    tw_core = line.twiss(co_guess={'x': 0.0}, num_turns=0)

    assert len(tw_mt.rows['_turn.*']) == 4
    assert len(tw_mt.rows['_end_poi.*']) == 1

    assert '_turn_0' in tw_mt.name
    assert '_turn_1' in tw_mt.name
    assert '_turn_2' in tw_mt.name
    assert '_turn_3' in tw_mt.name

    assert np.all(np.diff(tw_mt.s) >= 0)

    circum = line.get_length()

    xo.assert_allclose(tw_mt.s[-1], 4 * circum, atol=1e-10, rtol=0)
    xo.assert_allclose(tw_mt['s', '_turn_0'], 0, atol=1e-10, rtol=0)
    xo.assert_allclose(tw_mt['s', '_turn_1'], circum, atol=1e-10, rtol=0)
    xo.assert_allclose(tw_mt['s', '_turn_2'], 2 * circum, atol=1e-10, rtol=0)
    xo.assert_allclose(tw_mt['s', '_turn_3'], 3 * circum, atol=1e-10, rtol=0)
    xo.assert_allclose(tw_mt['s', '_end_point'], 4 * circum, atol=1e-10, rtol=0)

    xo.assert_allclose(tw_mt.mux[-1], 4 * tw.mux[-1], rtol=0, atol=0.05)
    xo.assert_allclose(tw_mt.muy[-1], 4 * tw.muy[-1], rtol=0, atol=0.05)

    assert 'qx' in tw
    assert 'qy' in tw
    assert 'qx' not in tw_mt
    assert 'qy' not in tw_mt

    assert tw_mt['x', '_turn_0'] > 2e-2
    assert np.abs(tw_mt['x', '_turn_1']) < 1e-2
    assert tw_mt['x', '_turn_2'] < -2e-2
    assert np.abs(tw_mt['x', '_turn_3']) < 1e-2

    xo.assert_allclose(tw_mt.x[-1], tw_mt.x[0], rtol=0, atol=1e-8)
    xo.assert_allclose(tw_mt.y[-1], tw_mt.y[0], rtol=0, atol=1e-8)
    xo.assert_allclose(tw_mt.px[-1], tw_mt.px[0], rtol=0, atol=1e-10)
    xo.assert_allclose(tw_mt.py[-1], tw_mt.py[0], rtol=0, atol=1e-10)
    xo.assert_allclose(tw_mt.betx[-1], tw_mt.betx[0], rtol=0, atol=1e-5)
    xo.assert_allclose(tw_mt.bety[-1], tw_mt.bety[0], rtol=0, atol=1e-5)
    xo.assert_allclose(tw_mt.alfx[-1], tw_mt.alfx[0], rtol=0, atol=1e-5)
    xo.assert_allclose(tw_mt.alfy[-1], tw_mt.alfy[0], rtol=0, atol=1e-5)


import pathlib

import numpy as np
import pytest
from cpymad.madx import Madx

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts
from xtrack.slicing import Strategy, Teapot

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@pytest.mark.parametrize('deferred_expressions', [True, False])
@for_all_test_contexts
def test_sps_thick(test_context, deferred_expressions):

    mad = Madx(stdout=False)
    mad.call(str(test_data_folder) + '/sps_thick/sps.seq')
    mad.input('beam, particle=proton, pc=26;')
    mad.call(str(test_data_folder) + '/sps_thick/lhc_q20.str')
    mad.use(sequence='sps')
    twmad = mad.twiss()

    line = xt.Line.from_madx_sequence(
        sequence=mad.sequence.sps,
        deferred_expressions=deferred_expressions,
        allow_thick=True)
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV,
                                    gamma0=mad.sequence.sps.beam.gamma)
    line.build_tracker(_context=test_context)
    line.twiss_default['method'] = '4d'

    # Check a bend
    assert line.element_names[26] == 'mbb.10150'

    assert isinstance(line['mbb.10150'], xt.RBend)

    assert line['mbb.10150'].edge_entry_model == 'linear'
    assert line['mbb.10150'].edge_exit_model == 'linear'
    assert line['mbb.10150'].model == 'adaptive'

    ang = line['mbb.10150'].k0 * line['mbb.10150'].length
    xo.assert_allclose(line['mbb.10150'].edge_entry_angle, 0, atol=1e-11, rtol=0)
    xo.assert_allclose(line['mbb.10150'].edge_exit_angle, 0, atol=1e-11, rtol=0)

    tw = line.twiss()
    xo.assert_allclose(twmad.s[-1], tw.s[-1], atol=1e-9, rtol=0)
    xo.assert_allclose(twmad.summary.q1, tw.qx, rtol=0, atol=1e-7)
    xo.assert_allclose(twmad.summary.q2, tw.qy, rtol=0, atol=1e-7)
    xo.assert_allclose(twmad.summary.dq1, tw.dqx, rtol=0, atol=0.2)
    xo.assert_allclose(twmad.summary.dq2, tw.dqy, rtol=0, atol=0.2)

    line.configure_bend_model(edge='full', core='full')

    tw = line.twiss()

    assert line['mbb.10150'].edge_entry_model == 'full'
    assert line['mbb.10150'].edge_exit_model == 'full'
    assert line['mbb.10150'].model == 'full'

    xo.assert_allclose(twmad.s[-1], tw.s[-1], atol=1e-9, rtol=0)
    xo.assert_allclose(twmad.summary.q1, tw.qx, rtol=0, atol=1e-7)
    xo.assert_allclose(twmad.summary.q2, tw.qy, rtol=0, atol=1e-7)
    xo.assert_allclose(twmad.summary.dq1, tw.dqx, rtol=0, atol=0.01)
    xo.assert_allclose(twmad.summary.dq2, tw.dqy, rtol=0, atol=0.01)

    line.configure_bend_model(core='expanded')

    tw = line.twiss()

    assert line['mbb.10150'].edge_entry_model == 'full'
    assert line['mbb.10150'].edge_exit_model == 'full'
    assert line['mbb.10150'].model == 'mat-kick-mat'

    xo.assert_allclose(twmad.s[-1], tw.s[-1], atol=1e-9, rtol=0)
    xo.assert_allclose(twmad.summary.q1, tw.qx, rtol=0, atol=1e-7)
    xo.assert_allclose(twmad.summary.q2, tw.qy, rtol=0, atol=1e-7)
    xo.assert_allclose(twmad.summary.dq1, tw.dqx, rtol=0, atol=0.2)
    xo.assert_allclose(twmad.summary.dq2, tw.dqy, rtol=0, atol=0.2)

    line.configure_bend_model(edge='linear')

    assert line['mbb.10150'].edge_entry_model == 'linear'
    assert line['mbb.10150'].edge_exit_model == 'linear'
    assert line['mbb.10150'].model == 'mat-kick-mat'

    line.configure_bend_model(core='full')
    line.configure_bend_model(edge='full')

    assert line['mbb.10150'].edge_entry_model == 'full'
    assert line['mbb.10150'].edge_exit_model == 'full'
    assert line['mbb.10150'].model == 'full'

    # Test from_dict/to_dict roundtrip
    dct = line.to_dict()
    line = xt.Line.from_dict(dct)

    assert line['mbb.10150'].edge_entry_model == 'full'
    assert line['mbb.10150'].edge_exit_model == 'full'
    assert line['mbb.10150'].model == 'full'

    line.discard_tracker()
    slicing_strategies = [
        Strategy(slicing=Teapot(1)),  # Default
        Strategy(slicing=Teapot(2), element_type=xt.Bend),
        Strategy(slicing=Teapot(2), element_type=xt.RBend),
        Strategy(slicing=Teapot(8), element_type=xt.Quadrupole),
    ]

    line.slice_thick_elements(slicing_strategies)
    line.build_tracker(_context=test_context)

    # Check a bend
    assert line.element_names[112] == 'mbb.10150_entry'
    assert line.element_names[113] == 'mbb.10150..entry_map'
    assert line.element_names[114] == 'drift_mbb.10150..0'
    assert line.element_names[115] == 'mbb.10150..0'
    assert line.element_names[116] == 'drift_mbb.10150..1'
    assert line.element_names[117] == 'mbb.10150..1'
    assert line.element_names[118] == 'drift_mbb.10150..2'
    assert line.element_names[119] == 'mbb.10150..exit_map'
    assert line.element_names[120] == 'mbb.10150_exit'

    assert isinstance(line['mbb.10150_entry'], xt.Marker)
    assert isinstance(line['mbb.10150..entry_map'], xt.ThinSliceRBendEntry)
    assert isinstance(line['drift_mbb.10150..0'], xt.DriftSliceRBend)
    assert isinstance(line['mbb.10150..0'], xt.ThinSliceRBend)
    assert isinstance(line['drift_mbb.10150..1'], xt.DriftSliceRBend)
    assert isinstance(line['mbb.10150..1'], xt.ThinSliceRBend)
    assert isinstance(line['drift_mbb.10150..2'], xt.DriftSliceRBend)
    assert isinstance(line['mbb.10150..exit_map'], xt.ThinSliceRBendExit)
    assert isinstance(line['mbb.10150_exit'], xt.Marker)

    # Check a quadrupole
    assert line.element_names[156] == 'qf.10210_entry'
    assert line.element_names[157] == 'qf.10210..entry_map'
    assert line.element_names[158] == 'drift_qf.10210..0'
    assert line.element_names[159] == 'qf.10210..0'
    assert line.element_names[160] == 'drift_qf.10210..1'
    assert line.element_names[161] == 'qf.10210..1'
    assert line.element_names[162] == 'drift_qf.10210..2'
    assert line.element_names[163] == 'qf.10210..2'
    assert line.element_names[164] == 'drift_qf.10210..3'
    assert line.element_names[165] == 'qf.10210..3'
    assert line.element_names[166] == 'drift_qf.10210..4'
    assert line.element_names[167] == 'qf.10210..4'
    assert line.element_names[168] == 'drift_qf.10210..5'
    assert line.element_names[169] == 'qf.10210..5'
    assert line.element_names[170] == 'drift_qf.10210..6'
    assert line.element_names[171] == 'qf.10210..6'
    assert line.element_names[172] == 'drift_qf.10210..7'
    assert line.element_names[173] == 'qf.10210..7'
    assert line.element_names[174] == 'drift_qf.10210..8'
    assert line.element_names[175] == 'qf.10210..exit_map'
    assert line.element_names[176] == 'qf.10210_exit'

    assert isinstance(line['qf.10210..7'], xt.ThinSliceQuadrupole)

    assert line['mbb.10150..entry_map']._parent.model == 'full'
    assert line['mbb.10150..exit_map']._parent.model == 'full'

    line.configure_bend_model(edge='linear')

    assert line['mbb.10150..entry_map']._parent.edge_entry_model == 'linear'
    assert line['mbb.10150..exit_map']._parent.edge_exit_model == 'linear'

    line.configure_bend_model(edge='full')

    assert line['mbb.10150..entry_map']._parent.edge_entry_model == 'full'
    assert line['mbb.10150..exit_map']._parent.edge_exit_model == 'full'

    tw_edge_full = line.twiss()

    xo.assert_allclose(twmad.s[-1], tw_edge_full.s[-1], atol=1e-9, rtol=0)
    xo.assert_allclose(twmad.summary.q1, tw_edge_full.qx, rtol=0, atol=0.5e-3)
    xo.assert_allclose(twmad.summary.q2, tw_edge_full.qy, rtol=0, atol=0.5e-3)
    xo.assert_allclose(twmad.summary.dq1, tw_edge_full.dqx, rtol=0, atol=0.2)
    xo.assert_allclose(twmad.summary.dq2, tw_edge_full.dqy, rtol=0, atol=0.2)

    line.configure_bend_model(edge='linear')
    tw_edge_linear = line.twiss()
    xo.assert_allclose(twmad.s[-1], tw_edge_linear.s[-1], atol=1e-9, rtol=0)
    xo.assert_allclose(twmad.summary.q1, tw_edge_linear.qx, rtol=0, atol=0.5e-3)
    xo.assert_allclose(twmad.summary.q2, tw_edge_linear.qy, rtol=0, atol=0.5e-3)
    xo.assert_allclose(twmad.summary.dq1, tw_edge_linear.dqx, rtol=0, atol=0.2)
    xo.assert_allclose(twmad.summary.dq2, tw_edge_linear.dqy, rtol=0, atol=0.2)

    tw_backwards = line.twiss(start=line.element_names[0],
                end=line.element_names[-1],
                init=tw_edge_linear.get_twiss_init(line.element_names[-1]),
                compute_chromatic_properties=True)

    assert_allclose = np.testing.assert_allclose

    assert_allclose(tw_backwards.s, tw_edge_linear.s, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.x, tw_edge_linear.x, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.px, tw_edge_linear.px, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.y, tw_edge_linear.y, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.py, tw_edge_linear.py, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.zeta, tw_edge_linear.zeta, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.delta, tw_edge_linear.delta, rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.betx, tw_edge_linear.betx, rtol=5e-9, atol=1e-10)
    assert_allclose(tw_backwards.bety, tw_edge_linear.bety, rtol=5e-9, atol=1e-10)
    assert_allclose(tw_backwards.ax_chrom, tw_edge_linear.ax_chrom, rtol=0, atol=1e-5)
    assert_allclose(tw_backwards.ay_chrom, tw_edge_linear.ay_chrom, rtol=0, atol=1e-5)
    assert_allclose(tw_backwards.bx_chrom, tw_edge_linear.bx_chrom, rtol=0, atol=1e-5)
    assert_allclose(tw_backwards.by_chrom, tw_edge_linear.by_chrom, rtol=0, atol=1e-5)
    assert_allclose(tw_backwards.dx, tw_edge_linear.dx, rtol=0, atol=1e-8)
    assert_allclose(tw_backwards.dy, tw_edge_linear.dy, rtol=0, atol=1e-8)
    assert_allclose(tw_backwards.mux[-1] - tw_edge_linear.mux[0], tw_edge_linear.qx,
                    rtol=0, atol=1e-10)
    assert_allclose(tw_backwards.muy[-1] - tw_edge_linear.muy[0], tw_edge_linear.qy,
                    rtol=0, atol=1e-10)



import numpy as np
import xtrack as xt
import xdeps as xd

from cpymad.madx import Madx

fname = 'fccee_z'; pc_gev = 45.6
# fname = 'fccee_w'; pc_gev = 80.
fname = 'fccee_h'; pc_gev = 120.
fname = 'fccee_t'; pc_gev = 182.5

mad = Madx()
mad.call('../../test_data/fcc_ee/' + fname + '.seq')
mad.beam(particle='positron', pc=pc_gev)
mad.use('fccee_p_ring')

mad.call('install_wigglers.madx')
mad.input("exec, define_wigglers_as_multipoles()")
mad.input("exec, install_wigglers()")
mad.use('fccee_p_ring')

line_thick = xt.Line.from_madx_sequence(mad.sequence.fccee_p_ring, allow_thick=True,
                                  deferred_expressions=True)

# Introduce 90 degree tilt for wiggler
tt = line_thick.get_table()
wigs = tt.rows['mwi.*', tt.element_type=='Bend'].name
for nn in wigs:
    line_thick.element_refs[nn].rot_s_rad = np.pi/2

line_thick.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                                 gamma0=mad.sequence.fccee_p_ring.beam.gamma)
line_thick.build_tracker()
tw_thick_no_rad = line_thick.twiss(method='4d')

line = line_thick.copy()
Strategy = xt.slicing.Strategy
Teapot = xt.slicing.Teapot
slicing_strategies = [
    Strategy(slicing=Teapot(1)),  # Default catch-all as in MAD-X
    Strategy(slicing=Teapot(3), element_type=xt.Bend),
    Strategy(slicing=Teapot(3), element_type=xt.CombinedFunctionMagnet),
    # Strategy(slicing=Teapot(50), element_type=xt.Quadrupole), # Starting point
    Strategy(slicing=Teapot(5), name=r'^qf.*'),
    Strategy(slicing=Teapot(5), name=r'^qd.*'),
    Strategy(slicing=Teapot(5), name=r'^qfg.*'),
    Strategy(slicing=Teapot(5), name=r'^qdg.*'),
    Strategy(slicing=Teapot(5), name=r'^ql.*'),
    Strategy(slicing=Teapot(5), name=r'^qs.*'),
    Strategy(slicing=Teapot(10), name=r'^qb.*'),
    Strategy(slicing=Teapot(10), name=r'^qg.*'),
    Strategy(slicing=Teapot(10), name=r'^qh.*'),
    Strategy(slicing=Teapot(10), name=r'^qi.*'),
    Strategy(slicing=Teapot(10), name=r'^qr.*'),
    Strategy(slicing=Teapot(10), name=r'^qu.*'),
    Strategy(slicing=Teapot(10), name=r'^qy.*'),
    Strategy(slicing=Teapot(50), name=r'^qa.*'),
    Strategy(slicing=Teapot(50), name=r'^qc.*'),
    Strategy(slicing=Teapot(20), name=r'^sy\..*'),
    Strategy(slicing=Teapot(30), name=r'^mwi\..*'),
]

line.slice_thick_elements(slicing_strategies=slicing_strategies)

line.build_tracker()
tw_thin_before = line.twiss(start=line.element_names[0], end=line.element_names[-1],
                          method='4d',
                          init=tw_thick_no_rad.get_twiss_init(0))

# Compare tunes
print('Before rematching:')

print('Tunes thick model:')
print(tw_thick_no_rad.qx, tw_thick_no_rad.qy)
print('Tunes thin model:')
print(tw_thin_before.mux[-1], tw_thin_before.muy[-1])

print('Beta beating at ips:')
print('H:', np.max(np.abs(
    tw_thin_before.rows['ip.*'].betx / tw_thick_no_rad.rows['ip.*'].betx -1)))
print('V:', np.max(np.abs(
    tw_thin_before.rows['ip.*'].bety / tw_thick_no_rad.rows['ip.*'].bety -1)))

print('Number of elements: ', len(line))
print('\n')

opt = line.match(
    method='4d',
    start=line.element_names[0], end=line.element_names[-1],
    init=tw_thick_no_rad.get_twiss_init(0),
    vary=xt.VaryList(['k1qf4', 'k1qf2', 'k1qd3', 'k1qd1',], step=1e-8,
    ),
    targets=[
        xt.TargetSet(at=xt.END, mux=tw_thick_no_rad.qx, muy=tw_thick_no_rad.qy, tol=1e-5),
    ]
)
opt.solve()
tw_thin_no_rad = line.twiss(method='4d')

print('After rematching:')
print('Tunes thick model:')
print(tw_thick_no_rad.qx, tw_thick_no_rad.qy)
print('Tunes thin model:')
print(tw_thin_no_rad.qx, tw_thin_no_rad.qy)

print('Beta beating at ips:')
print('H:', np.max(np.abs(
    tw_thin_no_rad.rows['ip.*'].betx / tw_thick_no_rad.rows['ip.*'].betx -1)))
print('V:', np.max(np.abs(
    tw_thin_no_rad.rows['ip.*'].bety / tw_thick_no_rad.rows['ip.*'].bety -1)))

print('Number of elements: ', len(line))

print('Change on arc quadrupoles:')
print(opt.log().vary[-1]/opt.log().vary[0] - 1)

print('\n Beta at the IPs:')
tw_thin_no_rad.rows['ip.*'].cols['betx bety'].show()

line.to_json(fname + '_thin.json')
line_thick.to_json(fname + '_thick.json')




from math import ceil
from typing import cast, Iterable, Collection

from .general import _print


class DefaultProgressIndicator:
    """Display progress of a task.

    This is a simple progress indicator, with an API that is a subset of the
    one of `tqdm`. It will provide simple progress information if tqdm is
    missing on the system.

    Parameters
    ----------
    iterable: Iterable
        The iterable to iterate over.
    desc: str
        Description of the task.
    total: int, optional
        Total number of iterations, if unspecified `len(iterable)` is used.
    miniters: int, optional
        Minimum number of iterations between updates, by default 1.
    unit_scale: int, optional
        Unused, kept for compatibility with tqdm. To be interpreted as the
        scale by which the iteration number and `total` are multiplied: i.e.,
        one iteration corresponds to `unit_scale` events.
    """
    def __init__(
            self,
            iterable: Iterable,
            desc: str,
            total: int = None,
            miniters: int = None,
            unit_scale: int = None,
    ):
        self.iterable = iterable
        self.desc = desc
        self._iterator = None
        self._iteration = 0
        self._total = total or len(cast(Collection, iterable))
        self._update_interval = miniters or ceil(self._total / 100)
        self._unit_scale = unit_scale or 1
        print(f'Init: interval {self._update_interval}, total {self._total}')

    def __iter__(self):
        self._iterator = iter(self.iterable)
        return self

    def __next__(self):
        if self._total and self._iteration % self._update_interval == 0:
            self._print_progress()

        try:
            return next(self._iterator)
        except StopIteration:
            self._print_progress()
            _print()  # finish with a newline
            raise
        finally:
            self._iteration += 1

    def _print_progress(self):
        if self._total == 0:
            # No meaningful progress over an empty collection
            _print(f'{self.desc}: done (0 iterations).', end='\r', flush=True)
            return

        percent = round(self._iteration * 100 / self._total)
        index = self._iteration * self._unit_scale
        scaled_total = self._total * self._unit_scale
        message = f'{self.desc}: {percent:3}% ({index}/{scaled_total} iterations)'
        _print(message, end='\r', flush=True)


class _ProgressIndicatorConfig:
    default_indicator_cls = DefaultProgressIndicator
    default_options = {}


_config = _ProgressIndicatorConfig()


def set_default_indicator(indicator_cls, **options):
    _config.default_indicator_cls = indicator_cls
    _config.default_options = options


def progress(iterable: Iterable, **options):
    indicator = _config.default_indicator_cls
    return indicator(iterable, **_config.default_options, **options)


try:
    import warnings
    with warnings.catch_warnings(): # To avoid warnings from tqdm.autonotebook (experimental module)
        warnings.simplefilter("ignore")
        from tqdm.autonotebook import tqdm
        from tqdm.notebook import tqdm_notebook

    if tqdm is tqdm_notebook:  # tqdm determined we're in a notebook.
        class PatchedTqdm(tqdm):
            """Patched version of tqdm_notebook.

            This patch is necessary, because notebook.tqdm has a bug. Until it
            is fixed, this workaround makes it possible to use tqdm with
            unit_scale in Jupyter.

            See issue tqdm/tqdm#1399 and PRs tqdm/tqdm#1528 and tqdm/tqdm#1461.
            """
            def __init__(self, *args, **kwargs):
                super(PatchedTqdm, self).__init__(*args, **kwargs)
                self.container.children[1].max = self.total

        tqdm = PatchedTqdm

    set_default_indicator(tqdm)
except ModuleNotFoundError:
    pass



import xtrack as xt

# Load a line and build a tracker
collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()

tw0 = collider.twiss(method='4d')

opt = collider.match(
    lines=['lhcb1', 'lhcb2'],
    start=['e.ds.l5.b1', 'e.ds.l5.b2'],
    end=['s.ds.r5.b1', 's.ds.r5.b2'],
    init=tw0,
    vary=xt.VaryList([
        'acbxv1.r5', 'acbxv1.l5', # <-- common elements
        'acbyvs4.l5b1', 'acbrdv4.r5b1', 'acbcv5.l5b1', # <-- b1
        'acbyvs4.l5b2', 'acbrdv4.r5b2', 'acbcv5.r5b2', # <-- b2
        ],
        step=1e-10, limits=[-1e-3, 1e-3]),
    targets = [
        xt.TargetSet(y=0, py=10e-6, at='ip5', line='lhcb1'),
        xt.TargetSet(y=0, py=-10e-6, at='ip5', line='lhcb2'),
        xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb1'),
        xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb2')
    ])
opt.target_status()

# prints:
#
# Target status:
# id state tag tol_met      residue  current_val target_val description
#  0 ON           True -3.93023e-19 -3.93023e-19          0 line=lhcb1, ('y', 'ip5'), val=0, tol=1e- ...
#  1 ON           True -7.06934e-18        1e-05      1e-05 line=lhcb1, ('py', 'ip5'), val=1e-05, to ...
#  2 ON           True -1.76183e-19 -1.76183e-19          0 line=lhcb2, ('y', 'ip5'), val=0, tol=1e- ...
#  3 ON           True -1.07353e-17       -1e-05     -1e-05 line=lhcb2, ('py', 'ip5'), val=-1e-05, t ...
#  4 ON           True  4.39323e-18  4.39323e-18          0 line=lhcb1, ('y', 's.ds.r5.b1'), val=0,  ...
#  5 ON           True  2.00777e-19  2.00777e-19          0 line=lhcb1, ('py', 's.ds.r5.b1'), val=0, ...
#  6 ON           True  5.23202e-19  5.23202e-19          0 line=lhcb2, ('y', 's.ds.r5.b2'), val=0,  ...
#  7 ON           True -4.05091e-20 -4.05091e-20          0 line=lhcb2, ('py', 's.ds.r5.b2'), val=0, ...

#!end-doc-part

import matplotlib.pyplot as plt

tw = collider.twiss()
plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.3, 4.8))
plt.plot(tw.lhcb1.s - tw.lhcb1['s', 'ip5'], tw.lhcb1.y*1000, label='b1', color='blue')
plt.plot(tw.lhcb2.s - tw.lhcb2['s', 'ip5'], tw.lhcb2.y*1000, label='b2', color='red')
plt.xlabel('s [m]')
plt.ylabel('y [mm]')
plt.legend()

for nn in ['mcbxfbv.a2r5', 'mcbxfbv.a2l5', 'ip5',
           'mcbyv.a4l5.b1', 'mcbrdv.4r5.b1', 'mcbcv.5l5.b1',
           'mcbyv.4l5.b2', 'mcbrdv.4r5.b2', 'mcbcv.5r5.b2']:
    if nn.endswith('b2'):
        lname = 'lhcb2'
        ha = 'right'
        color = 'red'
    elif nn.endswith('b1'):
        lname = 'lhcb1'
        ha = 'left'
        color = 'blue'
    elif nn.startswith('mcbx'):
        lname = 'lhcb1'
        ha = 'right'
        color = 'green'
    else:
        lname = 'lhcb1'
        ha = 'left'
        color = 'black'
    plt.axvline(x=tw[lname]['s', nn] - tw[lname]['s', 'ip5'], color=color, linestyle='--', alpha=0.5)
    plt.text(tw[lname]['s', nn] - tw[lname]['s', 'ip5'], -1.1, nn, rotation=90,
            horizontalalignment=ha, verticalalignment='bottom', color=color)

plt.ylim(-1.15, 0.6)
plt.xlim(-250, 250)

import numpy as np
assert np.isclose(tw.lhcb1['y', 'ip5'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb1['py', 'ip5'], 10e-6, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb1['y', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb1['py', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb2['y', 'ip5'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb2['py', 'ip5'], -10e-6, rtol=0, atol=1e-10)
assert np.isclose(tw.lhcb2['y', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)
assert np.isclose(tw.lhcb2['py', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)


plt.show()



import itertools
import numpy as np
import pytest

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

assert_allclose = xo.assert_allclose

@pytest.mark.parametrize('bend_type', [xt.Bend, xt.RBend])
@for_all_test_contexts
def test_thin_slice_bend(test_context, bend_type):

    bend = bend_type(k0=0.4, h=0.3, length=1,
                    edge_entry_angle=0.05, edge_entry_hgap=0.06, edge_entry_fint=0.08,
                    edge_exit_angle=0.05, edge_exit_hgap=0.06, edge_exit_fint=0.08)

    line = xt.Line(elements=[bend])

    line.configure_bend_model(edge='linear', core='expanded')

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(10000))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..995'].parent_name == 'e0'
    assert line['e0..995']._parent is line.element_dict['e0']
    assert line['drift_e0..995'].parent_name == 'e0'
    assert line['drift_e0..995']._parent is line.element_dict['e0']
    assert line['e0..entry_map'].parent_name == 'e0'
    assert line['e0..entry_map']._parent is line.element_dict['e0']
    assert line['e0..exit_map'].parent_name == 'e0'
    assert line['e0..exit_map']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    thin_slice_cls, drift_slice_type_cls, slice_entry_cls, slice_exit_cls = {
        xt.Bend: (
            xt.ThinSliceBend,
            xt.DriftSliceBend,
            xt.ThinSliceBendEntry,
            xt.ThinSliceBendExit,
        ),
        xt.RBend: (
            xt.ThinSliceRBend,
            xt.DriftSliceRBend,
            xt.ThinSliceRBendEntry,
            xt.ThinSliceRBendExit
        ),
    }[bend_type]

    line.to_json('ttt_thin_bend.json')
    line2 = xt.Line.from_json('ttt_thin_bend.json')
    assert isinstance(line2['e0..995'], thin_slice_cls)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is None
    assert line2['drift_e0..995'].parent_name == 'e0'
    assert line2['drift_e0..995']._parent is None
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is None
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..995'], thin_slice_cls)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..995'], drift_slice_type_cls)
    assert line2['drift_e0..995'].parent_name == 'e0'
    assert line2['drift_e0..995']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..entry_map'], slice_entry_cls)
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..exit_map'], slice_exit_cls)
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

    line.optimize_for_tracking()

    assert isinstance(line['e0..995'], xt.SimpleThinBend)
    assert isinstance(line['drift_e0..995'], xt.Drift)

    assert isinstance(line['e0..entry_map'], xt.DipoleEdge)
    assert isinstance(line['e0..exit_map'], xt.DipoleEdge)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thin_slice_quadrupole(test_context):

    quad = xt.Quadrupole(k1=0.1, length=1)

    line = xt.Line(elements=[quad])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(10000))])
    line.build_tracker(_context=test_context)
    assert line['e0..995'].parent_name == 'e0'
    assert line['e0..995']._parent is line.element_dict['e0']
    assert line['drift_e0..995'].parent_name == 'e0'
    assert line['drift_e0..995']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    quad.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thin_quad.json')
    line2 = xt.Line.from_json('ttt_thin_quad.json')
    assert isinstance(line2['e0..995'], xt.ThinSliceQuadrupole)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is None
    assert line2['drift_e0..995'].parent_name == 'e0'
    assert line2['drift_e0..995']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..995'], xt.ThinSliceQuadrupole)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..995'], xt.DriftSliceQuadrupole)
    assert line2['drift_e0..995'].parent_name == 'e0'
    assert line2['drift_e0..995']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

    line.optimize_for_tracking()

    assert isinstance(line['e0..995'], xt.SimpleThinQuadrupole)
    assert isinstance(line['drift_e0..995'], xt.Drift)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thin_slice_sextupole(test_context):

    sext = xt.Sextupole(k2=0.1, length=1,
                shift_x=1e-3, shift_y=2e-3, rot_s_rad=0.2
                )

    line = xt.Line(elements=[sext])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1))])
    line.build_tracker(_context=test_context)
    assert line['e0..0'].parent_name == 'e0'
    assert line['e0..0']._parent is line.element_dict['e0']
    assert line['drift_e0..0'].parent_name == 'e0'
    assert line['drift_e0..0']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    sext.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thin_sext.json')
    line2 = xt.Line.from_json('ttt_thin_sext.json')
    assert isinstance(line2['e0..0'], xt.ThinSliceSextupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is None
    assert line2['drift_e0..0'].parent_name == 'e0'
    assert line2['drift_e0..0']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThinSliceSextupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..0'], xt.DriftSliceSextupole)
    assert line2['drift_e0..0'].parent_name == 'e0'
    assert line2['drift_e0..0']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

    line.optimize_for_tracking()

    assert isinstance(line['e0..0'], xt.Multipole)
    assert isinstance(line['drift_e0..0'], xt.Drift)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thin_slice_octupole(test_context):

    oct = xt.Octupole(k3=0.1, length=1)

    line = xt.Line(elements=[oct])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1))])
    line.build_tracker(_context=test_context)
    assert line['e0..0'].parent_name == 'e0'
    assert line['e0..0']._parent is line.element_dict['e0']
    assert line['drift_e0..0'].parent_name == 'e0'
    assert line['drift_e0..0']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    oct.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thin_oct.json')
    line2 = xt.Line.from_json('ttt_thin_oct.json')
    assert isinstance(line2['e0..0'], xt.ThinSliceOctupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is None
    assert line2['drift_e0..0'].parent_name == 'e0'
    assert line2['drift_e0..0']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThinSliceOctupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..0'], xt.DriftSliceOctupole)
    assert line2['drift_e0..0'].parent_name == 'e0'
    assert line2['drift_e0..0']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

    line.optimize_for_tracking()

    assert isinstance(line['e0..0'], xt.Multipole)
    assert isinstance(line['drift_e0..0'], xt.Drift)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thin_slice_drift(test_context):

    drift = xt.Drift(length=1)

    line = xt.Line(elements=[drift])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(5), element_type=xt.Drift)])
    line.build_tracker(_context=test_context)
    assert line['drift_e0..0'].parent_name == 'e0'
    assert line['drift_e0..0']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                      _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    drift.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_drift.json')
    line2 = xt.Line.from_json('ttt_drift.json')
    assert line2['drift_e0..0'].parent_name == 'e0'
    assert line2['drift_e0..0']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['drift_e0..0'], xt.DriftSlice)
    assert line2['drift_e0..0'].parent_name == 'e0'
    assert line2['drift_e0..0']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

    line.optimize_for_tracking()

    assert isinstance(line['drift_e0..0'], xt.Drift)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)


@pytest.mark.parametrize('bend_type', [xt.Bend, xt.RBend])
@for_all_test_contexts
def test_thick_slice_bend(test_context, bend_type):

    bend = bend_type(k0=0.4, h=0.3, length=1,
                     edge_entry_angle=0.05, edge_entry_hgap=0.06, edge_entry_fint=0.08,
                     edge_exit_angle=0.05, edge_exit_hgap=0.06, edge_exit_fint=0.08)

    line = xt.Line(elements=[bend])

    line.configure_bend_model(edge='linear', core='expanded')

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(5, mode='thick'))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..3'].parent_name == 'e0'
    assert line['e0..3']._parent is line.element_dict['e0']
    assert line['e0..entry_map'].parent_name == 'e0'
    assert line['e0..entry_map']._parent is line.element_dict['e0']
    assert line['e0..exit_map'].parent_name == 'e0'
    assert line['e0..exit_map']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    thick_slice_cls, thick_slice_bend_cls = {
        xt.Bend: (
            xt.ThickSliceBend,
            xt.ThickSliceBend
        ),
        xt.RBend: (
            xt.ThickSliceRBend,
            xt.ThickSliceRBend
        ),
    }[bend_type]

    line.to_json('ttt_thick_bend.json')
    line2 = xt.Line.from_json('ttt_thick_bend.json')
    assert isinstance(line2['e0..3'], thick_slice_cls)
    assert line2['e0..3'].parent_name == 'e0'
    assert line2['e0..3']._parent is None
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is None
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..3'], thick_slice_bend_cls)
    assert line2['e0..3'].parent_name == 'e0'
    assert line2['e0..3']._parent is line2.element_dict['e0']
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is line2.element_dict['e0']
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thick_slice_quadrupole(test_context):

    quad = xt.Quadrupole(k1=0.1, length=1)

    line = xt.Line(elements=[quad])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(5, mode='thick'))])
    line.build_tracker(_context=test_context)
    assert line['e0..3'].parent_name == 'e0'
    assert line['e0..3']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    quad.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thick_quad.json')
    line2 = xt.Line.from_json('ttt_thick_quad.json')
    assert isinstance(line2['e0..3'], xt.ThickSliceQuadrupole)
    assert line2['e0..3'].parent_name == 'e0'
    assert line2['e0..3']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..3'], xt.ThickSliceQuadrupole)
    assert line2['e0..3'].parent_name == 'e0'
    assert line2['e0..3']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thick_slice_sextupole(test_context):

    sext = xt.Sextupole(k2=0.1, length=1)

    line = xt.Line(elements=[sext])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1, mode='thick'))])
    line.build_tracker(_context=test_context)
    assert line['e0..0'].parent_name == 'e0'
    assert line['e0..0']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    sext.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thick_sext.json')
    line2 = xt.Line.from_json('ttt_thick_sext.json')
    assert isinstance(line2['e0..0'], xt.ThickSliceSextupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThickSliceSextupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thick_slice_octupole(test_context):

    oct = xt.Octupole(k3=0.1, length=1)

    line = xt.Line(elements=[oct])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1, mode='thick'))])
    line.build_tracker(_context=test_context)
    assert line['e0..0'].parent_name == 'e0'
    assert line['e0..0']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    oct.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thick_oct.json')
    line2 = xt.Line.from_json('ttt_thick_oct.json')
    assert isinstance(line2['e0..0'], xt.ThickSliceOctupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThickSliceOctupole)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thick_slice_solenoid(test_context):

    sol = xt.Solenoid(ks=0.1, length=1)

    line = xt.Line(elements=[sol])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(10, mode='thick'))])
    line.build_tracker(_context=test_context)
    assert line['e0..0'].parent_name == 'e0'
    assert line['e0..0']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    sol.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

    line.to_json('ttt_thick_solenoid.json')
    line2 = xt.Line.from_json('ttt_thick_solenoid.json')
    assert isinstance(line2['e0..0'], xt.ThickSliceSolenoid)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThickSliceSolenoid)
    assert line2['e0..0'].parent_name == 'e0'
    assert line2['e0..0']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)


@for_all_test_contexts
@pytest.mark.parametrize(
    'model',
    ['full', 'linear'])
def test_bend_edge_slice_entry(test_context, model):

    # only edge entry
    bend_only_e1 = xt.Bend(
        length=0, k0=0.1,
        edge_entry_angle=0.05,
        edge_entry_hgap=0.06,
        edge_entry_fint=0.08,
        edge_exit_active=False,
        _context=test_context)

    edge_e1 = xt.DipoleEdge(
        k=0.1, side='entry', e1=0.05, hgap=0.06, fint=0.08,
        _context=test_context)

    line = xt.Line(elements=[bend_only_e1])

    line['e0'].length = 1 # to force the slicing
    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1))])
    line['e0'].length = 0
    line.build_tracker(_context=test_context)

    assert 'e0..entry_map' in line.element_names
    assert 'e0..exit_map' in line.element_names
    assert isinstance(line['e0..entry_map'], xt.ThinSliceBendEntry)
    assert isinstance(line['e0..exit_map'], xt.ThinSliceBendExit)

    edge_e1.model = model
    bend_only_e1.edge_entry_model = model

    p1 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                      _context=test_context)
    p2 = p1.copy()
    p3 = p1.copy()

    bend_only_e1.track(p1)
    edge_e1.track(p2)
    line.track(p3)

    assert_allclose(p1.x, p2.x, rtol=0, atol=1e-14)
    assert_allclose(p1.px, p2.px, rtol=0, atol=1e-14)
    assert_allclose(p1.y, p2.y, rtol=0, atol=1e-14)
    assert_allclose(p1.py, p2.py, rtol=0, atol=1e-14)
    assert_allclose(p1.zeta, p2.zeta, rtol=0, atol=1e-14)
    assert_allclose(p1.delta, p2.delta, rtol=0, atol=1e-14)

    assert_allclose(p1.x, p3.x, rtol=0, atol=1e-14)
    assert_allclose(p1.px, p3.px, rtol=0, atol=1e-14)
    assert_allclose(p1.y, p3.y, rtol=0, atol=1e-14)
    assert_allclose(p1.py, p3.py, rtol=0, atol=1e-14)
    assert_allclose(p1.zeta, p3.zeta, rtol=0, atol=1e-14)
    assert_allclose(p1.delta, p3.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
@pytest.mark.parametrize(
    'model',
    ['full', 'linear'])
def test_bend_edge_slice_exit(test_context, model):

    # only edge entry
    bend_only_e1 = xt.Bend(
        length=0, k0=0.1,
        edge_entry_angle=0.05,
        edge_entry_hgap=0.06,
        edge_entry_fint=0.08,
        edge_exit_active=False,
        _context=test_context)

    edge_e1 = xt.DipoleEdge(
        k=0.1, side='entry', e1=0.05, hgap=0.06, fint=0.08, _context=test_context)

    line = xt.Line(elements=[bend_only_e1])

    line['e0'].length = 1 # to force the slicing
    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1))])
    line['e0'].length = 0
    line.build_tracker(_context=test_context)

    assert 'e0..entry_map' in line.element_names
    assert 'e0..exit_map' in line.element_names
    assert isinstance(line['e0..entry_map'], xt.ThinSliceBendEntry)
    assert isinstance(line['e0..exit_map'], xt.ThinSliceBendExit)

    edge_e1.model = model
    bend_only_e1.edge_entry_model = model

    p1 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                      _context=test_context)
    p2 = p1.copy()
    p3 = p1.copy()

    bend_only_e1.track(p1)
    edge_e1.track(p2)
    line.track(p3)

    assert_allclose(p1.x, p2.x, rtol=0, atol=1e-14)
    assert_allclose(p1.px, p2.px, rtol=0, atol=1e-14)
    assert_allclose(p1.y, p2.y, rtol=0, atol=1e-14)
    assert_allclose(p1.py, p2.py, rtol=0, atol=1e-14)
    assert_allclose(p1.zeta, p2.zeta, rtol=0, atol=1e-14)
    assert_allclose(p1.delta, p2.delta, rtol=0, atol=1e-14)

    assert_allclose(p1.x, p3.x, rtol=0, atol=1e-14)
    assert_allclose(p1.px, p3.px, rtol=0, atol=1e-14)
    assert_allclose(p1.y, p3.y, rtol=0, atol=1e-14)
    assert_allclose(p1.py, p3.py, rtol=0, atol=1e-14)
    assert_allclose(p1.zeta, p3.zeta, rtol=0, atol=1e-14)
    assert_allclose(p1.delta, p3.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
def test_thin_slice_bend_with_multipoles(test_context):

    bend = xt.Bend(k0=0.4, h=0.3, length=1,
                   k1=0.003,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7],
                   num_multipole_kicks=1000,
                   edge_entry_angle=0.05, edge_entry_hgap=0.06, edge_entry_fint=0.08,
                   edge_exit_angle=0.05, edge_exit_hgap=0.06, edge_exit_fint=0.08)

    line = xt.Line(elements=[bend])
    bend.integrator = 'teapot'
    bend.model = 'drift-kick-drift-expanded'

    line.configure_bend_model(edge='linear')

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(1000))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..995'].parent_name == 'e0'
    assert line['e0..995']._parent is line.element_dict['e0']
    assert line['drift_e0..995'].parent_name == 'e0'
    assert line['drift_e0..995']._parent is line.element_dict['e0']
    assert line['e0..entry_map'].parent_name == 'e0'
    assert line['e0..entry_map']._parent is line.element_dict['e0']
    assert line['e0..exit_map'].parent_name == 'e0'
    assert line['e0..exit_map']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_bend_mult.json')
    line2 = xt.Line.from_json('ttt_bend_mult.json')
    assert isinstance(line2['e0..995'], xt.ThinSliceBend)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is None
    assert line2['drift_e0..995'].parent_name == 'e0'
    assert line2['drift_e0..995']._parent is None
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is None
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..995'], xt.ThinSliceBend)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..995'], xt.DriftSliceBend)
    assert line2['drift_e0..995'].parent_name == 'e0'
    assert line2['drift_e0..995']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..entry_map'], xt.ThinSliceBendEntry)
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..exit_map'], xt.ThinSliceBendExit)
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

    line.optimize_for_tracking()

    assert isinstance(line['e0..995'], xt.Multipole)
    assert isinstance(line['drift_e0..995'], xt.Drift)

    assert isinstance(line['e0..entry_map'], xt.DipoleEdge)
    assert isinstance(line['e0..exit_map'], xt.DipoleEdge)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thick_slice_bend_with_multipoles(test_context):

    bend = xt.Bend(k0=0.4, h=0.3, length=1,
                   k1=0.003,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7],
                   num_multipole_kicks=100000,
                   edge_entry_angle=0.05, edge_entry_hgap=0.06, edge_entry_fint=0.08,
                   edge_exit_angle=0.05, edge_exit_hgap=0.06, edge_exit_fint=0.08)

    line = xt.Line(elements=[bend])

    line.configure_bend_model(edge='linear', core='expanded')

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(10000, mode='thick'))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..995'].parent_name == 'e0'
    assert line['e0..995']._parent is line.element_dict['e0']
    assert line['e0..entry_map'].parent_name == 'e0'
    assert line['e0..entry_map']._parent is line.element_dict['e0']
    assert line['e0..exit_map'].parent_name == 'e0'
    assert line['e0..exit_map']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_thick_bend_mult.json')
    line2 = xt.Line.from_json('ttt_thick_bend_mult.json')
    assert isinstance(line2['e0..995'], xt.ThickSliceBend)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is None
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is None
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..995'], xt.ThickSliceBend)
    assert line2['e0..995'].parent_name == 'e0'
    assert line2['e0..995']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..entry_map'], xt.ThinSliceBendEntry)
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..exit_map'], xt.ThinSliceBendExit)
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

@for_all_test_contexts
def test_thin_slice_bend_with_multipoles_bend_off(test_context):

    num_slices = 10

    bend = xt.Bend(k0=0, h=0, length=1,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7],
                   num_multipole_kicks=num_slices,
                   edge_entry_angle=0.05, edge_entry_hgap=0.06, edge_entry_fint=0.08,
                   edge_exit_angle=0.05, edge_exit_hgap=0.06, edge_exit_fint=0.08)
    bend.integrator = 'teapot'

    line = xt.Line(elements=[bend])

    line.configure_bend_model(edge='linear', core='expanded')


    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(num_slices))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..5'].parent_name == 'e0'
    assert line['e0..5']._parent is line.element_dict['e0']
    assert line['drift_e0..5'].parent_name == 'e0'
    assert line['drift_e0..5']._parent is line.element_dict['e0']
    assert line['e0..entry_map'].parent_name == 'e0'
    assert line['e0..entry_map']._parent is line.element_dict['e0']
    assert line['e0..exit_map'].parent_name == 'e0'
    assert line['e0..exit_map']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-14)

    line.to_json('ttt_thin_bend_mult_off.json')
    line2 = xt.Line.from_json('ttt_thin_bend_mult_off.json')
    assert isinstance(line2['e0..5'], xt.ThinSliceBend)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is None
    assert line2['drift_e0..5'].parent_name == 'e0'
    assert line2['drift_e0..5']._parent is None
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is None
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..5'], xt.ThinSliceBend)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..5'], xt.DriftSliceBend)
    assert line2['drift_e0..5'].parent_name == 'e0'
    assert line2['drift_e0..5']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..entry_map'], xt.ThinSliceBendEntry)
    assert line2['e0..entry_map'].parent_name == 'e0'
    assert line2['e0..entry_map']._parent is line2.element_dict['e0']
    assert isinstance(line2['e0..exit_map'], xt.ThinSliceBendExit)
    assert line2['e0..exit_map'].parent_name == 'e0'
    assert line2['e0..exit_map']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)

    line.optimize_for_tracking()

    assert_allclose(line['e0..5'].knl[5], 0.6/num_slices, rtol=0, atol=1e-14)
    assert_allclose(line['e0..5'].ksl[5], 0.7/num_slices, rtol=0, atol=1e-14)

    assert isinstance(line['e0..5'], xt.Multipole)
    assert isinstance(line['drift_e0..5'], xt.Drift)

    assert isinstance(line['e0..entry_map'], xt.DipoleEdge)
    assert isinstance(line['e0..exit_map'], xt.DipoleEdge)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-14)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
def test_thick_slice_quad_with_multipoles(test_context):

    quad = xt.Quadrupole(k1=1e-3, k1s=2e-3, length=1,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7],
                   num_multipole_kicks=1000)

    line = xt.Line(elements=[quad])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(1000, mode='thick'))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..5'].parent_name == 'e0'
    assert line['e0..5']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_thick_quad_mult.json')
    line2 = xt.Line.from_json('ttt_thick_quad_mult.json')
    assert isinstance(line2['e0..5'], xt.ThickSliceQuadrupole)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..5'], xt.ThickSliceQuadrupole)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-8)

@for_all_test_contexts
def test_thin_slice_quad_with_multipoles(test_context):

    quad = xt.Quadrupole(k1=1e-3, k1s=2e-3, length=1,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7],
                   num_multipole_kicks=7)
    quad.integrator = 'teapot'
    quad.model = 'drift-kick-drift-expanded'

    line = xt.Line(elements=[quad])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(7))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..5'].parent_name == 'e0'
    assert line['e0..5']._parent is line.element_dict['e0']
    assert line['drift_e0..5'].parent_name == 'e0'
    assert line['drift_e0..5']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_thin_quad_mult.json')
    line2 = xt.Line.from_json('ttt_thin_quad_mult.json')
    assert isinstance(line2['e0..5'], xt.ThinSliceQuadrupole)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is None
    assert line2['drift_e0..5'].parent_name == 'e0'
    assert line2['drift_e0..5']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..5'], xt.ThinSliceQuadrupole)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..5'], xt.DriftSliceQuadrupole)
    assert line2['drift_e0..5'].parent_name == 'e0'
    assert line2['drift_e0..5']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-8)

    line.optimize_for_tracking()

    assert isinstance(line['e0..5'], xt.Multipole)
    assert isinstance(line['drift_e0..5'], xt.Drift)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
def test_thin_slice_quad_with_multipoles_quad_off(test_context):

    quad = xt.Quadrupole(k1=0, k1s=0, length=1,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7],
                   num_multipole_kicks=7)
    quad.model = 'drift-kick-drift-expanded'
    quad.integrator = 'teapot'

    line = xt.Line(elements=[quad])

    num_slices = 7

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Teapot(num_slices))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..5'].parent_name == 'e0'
    assert line['e0..5']._parent is line.element_dict['e0']
    assert line['drift_e0..5'].parent_name == 'e0'
    assert line['drift_e0..5']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.5, y=0.3, py=0.3, delta=0.1
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-14)

    line.to_json('ttt_thin_mult_quad_off.json')
    line2 = xt.Line.from_json('ttt_thin_mult_quad_off.json')
    assert isinstance(line2['e0..5'], xt.ThinSliceQuadrupole)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is None
    assert line2['drift_e0..5'].parent_name == 'e0'
    assert line2['drift_e0..5']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..5'], xt.ThinSliceQuadrupole)
    assert line2['e0..5'].parent_name == 'e0'
    assert line2['e0..5']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..5'], xt.DriftSliceQuadrupole)
    assert line2['drift_e0..5'].parent_name == 'e0'
    assert line2['drift_e0..5']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)

    line.optimize_for_tracking()

    assert isinstance(line['e0..5'], xt.Multipole)
    assert isinstance(line['drift_e0..5'], xt.Drift)

    assert_allclose(line['e0..5'].knl[5], 0.6/num_slices, rtol=0, atol=1e-14)
    assert_allclose(line['e0..5'].ksl[5], 0.7/num_slices, rtol=0, atol=1e-14)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-14)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
def test_thick_slice_sextupole_with_multipoles(test_context):

    sext = xt.Sextupole(k2=1e-3, k2s=2e-3,
                   length=0.001, # need to make it very short because thick has only one kick in the center
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7]
    )

    line = xt.Line(elements=[sext])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(2, mode='thick'))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..1'].parent_name == 'e0'
    assert line['e0..1']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_thick_sext_mult.json')
    line2 = xt.Line.from_json('ttt_thick_sext_mult.json')
    assert isinstance(line2['e0..0'], xt.ThickSliceSextupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThickSliceSextupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-8)

@for_all_test_contexts
def test_thin_slice_sextupole_with_multipoles(test_context):

    sext = xt.Sextupole(k2=1e-3, k2s=2e-3, length=0.001,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7])

    line = xt.Line(elements=[sext])

    num_slices = 2

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(num_slices))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..1'].parent_name == 'e0'
    assert line['e0..1']._parent is line.element_dict['e0']
    assert line['drift_e0..1'].parent_name == 'e0'
    assert line['drift_e0..1']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_thin_sext_mult.json')
    line2 = xt.Line.from_json('ttt_thin_sext_mult.json')
    assert isinstance(line2['e0..1'], xt.ThinSliceSextupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is None
    assert line2['drift_e0..1'].parent_name == 'e0'
    assert line2['drift_e0..1']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..1'], xt.ThinSliceSextupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..1'], xt.DriftSliceSextupole)
    assert line2['drift_e0..1'].parent_name == 'e0'
    assert line2['drift_e0..1']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-8)

    line.optimize_for_tracking()

    assert isinstance(line['e0..1'], xt.Multipole)
    assert isinstance(line['drift_e0..1'], xt.Drift)

    assert_allclose(line['e0..1'].knl[5], 0.6/num_slices, rtol=0, atol=1e-14)
    assert_allclose(line['e0..1'].ksl[5], 0.7/num_slices, rtol=0, atol=1e-14)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
def test_thick_slice_octupole_with_multipoles(test_context):

    oct = xt.Octupole(k3=1e-3, k3s=2e-3,
                   length=0.001, # need to make it very short because thick has only one kick in the center
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7])

    line = xt.Line(elements=[oct])

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(2, mode='thick'))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..1'].parent_name == 'e0'
    assert line['e0..1']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_thick_oct_mult.json')
    line2 = xt.Line.from_json('ttt_thick_oct_mult.json')
    assert isinstance(line2['e0..0'], xt.ThickSliceOctupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..0'], xt.ThickSliceOctupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-8)


@for_all_test_contexts
def test_thin_slice_octupole_with_multipoles(test_context):

    oct = xt.Octupole(k3=1e-3, k3s=-1e-3, length=0.001,
                   knl=[0, 0.001, 0.01, 0.02, 0.04, 0.6],
                   ksl=[0, 0.002, 0.03, 0.03, 0.05, 0.7])

    line = xt.Line(elements=[oct])

    num_slices = 2

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(num_slices))])
    line.build_tracker(_context=test_context)
    line._line_before_slicing.build_tracker(_context=test_context)
    assert line['e0..1'].parent_name == 'e0'
    assert line['e0..1']._parent is line.element_dict['e0']
    assert line['drift_e0..1'].parent_name == 'e0'
    assert line['drift_e0..1']._parent is line.element_dict['e0']

    p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03
                      ,_context=test_context)
    p_ref = p0.copy()
    p_slice = p0.copy()

    line.track(p_slice)
    line._line_before_slicing.track(p_ref)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.to_json('ttt_octupole_thin.json')
    line2 = xt.Line.from_json('ttt_octupole_thin.json')
    assert isinstance(line2['e0..1'], xt.ThinSliceOctupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is None
    assert line2['drift_e0..1'].parent_name == 'e0'
    assert line2['drift_e0..1']._parent is None

    line2.build_tracker(_context=test_context)
    assert isinstance(line2['e0..1'], xt.ThinSliceOctupole)
    assert line2['e0..1'].parent_name == 'e0'
    assert line2['e0..1']._parent is line2.element_dict['e0']
    assert isinstance(line2['drift_e0..1'], xt.DriftSliceOctupole)
    assert line2['drift_e0..1'].parent_name == 'e0'
    assert line2['drift_e0..1']._parent is line2.element_dict['e0']

    line.track(p_slice, backtrack=True)

    assert (p_slice.state == 1).all()
    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-8)

    line.optimize_for_tracking()

    assert isinstance(line['e0..1'], xt.Multipole)
    assert isinstance(line['drift_e0..1'], xt.Drift)

    assert_allclose(line['e0..1'].knl[5], 0.6/num_slices, rtol=0, atol=1e-14)
    assert_allclose(line['e0..1'].ksl[5], 0.7/num_slices, rtol=0, atol=1e-14)

    p_slice = p0.copy()
    line.track(p_slice)

    assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-8)
    assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-8)
    assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-8)
    assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-8)
    assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-8)
    assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-8)

    line.track(p_slice, backtrack=True)

    assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-14)
    assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-14)
    assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-14)
    assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-14)
    assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-14)
    assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-14)


@pytest.mark.parametrize('element_class', ['quadrupole', 'octupole', 'sextupole'])
def test_sliced_magnet_fringes(element_class):
    env = xt.Environment()

    if element_class == 'quadrupole':
        element = env.new('qd', xt.Quadrupole, k1=0.1, length=0.1)
    elif element_class == 'sextupole':
        element = env.new('sd', xt.Sextupole, k2=0.1, length=0.1)
    elif element_class == 'octupole':
        element = env.new('od', xt.Octupole, k3=0.1, length=0.1)

    line = env.new_line(name='line', components=[element])
    line_thick = line.copy(shallow=True)
    line.slice_thick_elements(slicing_strategies=[xt.Strategy(slicing=xt.Teapot(3))])

    p0 = xt.Particles(kinetic_energy0=100e6, mass0=xt.PROTON_MASS_EV,
                      x=1e-3, px=1e-3, y=1e-3, py=1e-3, zeta=1e-3, delta=1e-3)

    p_ref_no_fringe = p0.copy()
    p_slice_no_fringe = p0.copy()

    line_thick.track(p_ref_no_fringe)
    line.track(p_slice_no_fringe)

    line.configure_quadrupole_model(edge='full')
    p_ref_fringe = p0.copy()
    p_slice_fringe = p0.copy()

    line_thick.track(p_ref_fringe)
    line.track(p_slice_fringe)

    diff_no_fringe_x = p_ref_no_fringe.x - p_slice_no_fringe.x
    diff_fringe_x = p_ref_fringe.x - p_slice_fringe.x

    xo.assert_allclose(diff_no_fringe_x, diff_fringe_x, rtol=0, atol=1e-16)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

context = xo.ContextCpu()

######################################
# Create a dummy collimation process #
######################################

class DummyInteractionProcess:

    def __init__(self, fraction_lost, fraction_secondary):

        self.fraction_lost = fraction_lost
        self.fraction_secondary = fraction_secondary

    def interact(self, particles):

        n_part = particles._num_active_particles

        # Kill some particles
        mask_kill = np.random.uniform(size=n_part) < self.fraction_lost
        particles.state[:n_part][mask_kill] = 0


        # Generate some more particles
        mask_secondary = np.random.uniform(size=n_part) < self.fraction_secondary
        n_products = np.sum(mask_secondary)
        if n_products>0:
            products = {
                's': particles.s[:n_part][mask_secondary],
                'x': particles.x[:n_part][mask_secondary],
                'px': particles.px[:n_part][mask_secondary],
                'y': particles.y[:n_part][mask_secondary],
                'py': particles.py[:n_part][mask_secondary],
                'zeta': particles.zeta[:n_part][mask_secondary],
                'delta': particles.delta[:n_part][mask_secondary],

                'mass_ratio': particles.x[:n_part][mask_secondary] *0 + .5,
                'charge_ratio': particles.x[:n_part][mask_secondary] *0 + .5,

                'parent_particle_id': particles.particle_id[:n_part][mask_secondary],
                'at_element': particles.at_element[:n_part][mask_secondary],
                'at_turn': particles.at_turn[:n_part][mask_secondary],
                }
        else:
            products = None

        return products

#############################################
# Create the corresponding beam interaction #
#############################################

beam_interaction = xt.BeamInteraction(
        interaction_process=DummyInteractionProcess(fraction_lost=0.1,
                                                    fraction_secondary=0.2))

############################################
# Go through the collimator multiple times #
############################################

particles = xt.Particles(_capacity=200,
        p0c=7000, x=np.linspace(-1e-3, 1e-3, 10))

for _ in range(10):
    beam_interaction.track(particles)

###############
# Some checks #
###############

assert particles._num_lost_particles >= 0
assert particles._num_active_particles >= 0

n_all_parts = particles._num_active_particles + particles._num_lost_particles

assert np.all(np.diff(particles.state) <= 0) # checks that there is no lost after active

# Check each id is present only once
ids = particles.particle_id[:n_all_parts]
assert len(list(set(ids))) == n_all_parts

# Check parent and secondaries have the same position
ind_secondaries = np.where(particles.parent_particle_id[:particles._num_active_particles] !=
                    particles.particle_id[:particles._num_active_particles])[0]
for ii in ind_secondaries:
    parent_id = particles.parent_particle_id[ii]
    parent_x = particles.x[np.where(particles.particle_id == parent_id)[0][0]]

    assert parent_x == particles.x[ii]



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt
import xobjects as xo

# In this case the element internal record is made of two separate tables each
# with its own record index.

class Table1(xo.HybridClass):
    _xofields = {
        '_index': xt.RecordIndex,
        'particle_x': xo.Float64[:],
        'particle_px': xo.Float64[:],
        'at_element': xo.Int64[:],
        'at_turn': xo.Int64[:],
        'particle_id': xo.Int64[:]
        }

class Table2(xo.HybridClass):
    _xofields = {
        '_index': xt.RecordIndex,
        'generated_rr': xo.Float64[:],
        'at_element': xo.Int64[:],
        'at_turn': xo.Int64[:],
        'particle_id': xo.Int64[:]
        }

class TestElementRecord(xo.HybridClass):
    _xofields = {
        'table1': Table1,
        'table2': Table2
        }

# The two tables in the internal record can be accessed independently in the C
# code of the beam element.

TestElement_track_method_source = r'''
    /*gpufun*/
    void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

        // Extract the record and record_index
        TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
        Table1Data table1 = NULL;
        Table2Data table2 = NULL;
        RecordIndex table1_index = NULL;
        RecordIndex table2_index = NULL;
        if (record){
            table1 = TestElementRecordData_getp_table1(record);
            table2 = TestElementRecordData_getp_table2(record);
            table1_index = Table1Data_getp__index(table1);
            table2_index = Table2Data_getp__index(table2);
        }

        int64_t n_kicks = TestElementData_get_n_kicks(el);
        // printf("n_kicks %d\n", (int)n_kicks);

        //start_per_particle_block (part0->part)

            // Record in table1 info about the ingoing particle
            if (record){
                // Get a slot in table1
                int64_t i_slot = RecordIndex_get_slot(table1_index);
                // The returned slot id is negative if record is NULL or if record is full
                //printf("i_slot %d\n", (int)i_slot);

                if (i_slot>=0){
                        Table1Data_set_particle_x(table1, i_slot,
                                                    LocalParticle_get_x(part));
                        Table1Data_set_particle_px(table1, i_slot,
                                                    LocalParticle_get_px(part));
                        Table1Data_set_at_element(table1, i_slot,
                                                    LocalParticle_get_at_element(part));
                        Table1Data_set_at_turn(table1, i_slot,
                                                    LocalParticle_get_at_turn(part));
                        Table1Data_set_particle_id(table1, i_slot,
                                                    LocalParticle_get_particle_id(part));
                }
            }

            for (int64_t i = 0; i < n_kicks; i++) {
                double rr = 1e-6 * RandomUniform_generate(part);
                LocalParticle_add_to_px(part, rr);

                // Record in table2 info about the generated kicks
                if (record){
                    // Get a slot in table2
                    int64_t i_slot = RecordIndex_get_slot(table2_index);
                    // The returned slot id is negative if record is NULL or if record is full
                    //printf("i_slot %d\n", (int)i_slot);

                    if (i_slot>=0){
                            Table2Data_set_generated_rr(table2, i_slot, rr);
                            Table2Data_set_at_element(table2, i_slot,
                                                        LocalParticle_get_at_element(part));
                            Table2Data_set_at_turn(table2, i_slot,
                                                        LocalParticle_get_at_turn(part));
                            Table2Data_set_particle_id(table2, i_slot,
                                                        LocalParticle_get_particle_id(part));
                    }
                }
            }

        //end_per_particle_block
    }
    '''

# To allow elements of a given type to store data in a structure of the type defined
# above we need to add in the element class an attribute called
# `_internal_record_class` to which we bind the data structure type defined above.

class TestElement(xt.BeamElement):
    _xofields={
        'n_kicks': xo.Int64,
        }
    _internal_record_class = TestElementRecord

    _depends_on = [xt.RandomUniform]

    _extra_c_sources = [TestElement_track_method_source]

# Once these steps are done, the TestElement and its recording feature are ready
# and can be used as follows.

# Beam elements can be created as usual
test_element1 = TestElement(n_kicks=10)
test_element2 = TestElement(n_kicks=5)

# The recording is by default disabled and can be enabled using the following
# dedicated function. The capacity allocated for the two tables needs to be
# provided in a dictionary:

io_buffer = xt.new_io_buffer()
xt.start_internal_logging(elements=[test_element1, test_element2],
                          io_buffer=io_buffer,
                          capacity={'table1':1000, 'table2':500})

# Make some particles
part = xt.Particles(p0c=6.5e12, x=[1e-3,2e-3,3e-3])
part._init_random_number_generator()

# Track through the first element
test_element1.track(part)
# Track through the second element
test_element2.track(part)

# We can now inspect the two tables in the `record`, e.g `record.table1.particle_x`,
# `record.table2.generated_rr`. The number of used slots in each can be found in
# the corresponding index object e.g. record.table1._index.num_recorded.

# The recording can be stopped with the following method:
xt.stop_internal_logging(elements=[test_element1, test_element2])

# Track more times (without logging information in `record`)
test_element1.track(part)
test_element2.track(part)

# The record can be accessed from test_element1.record and test_element2.record

#!end-doc-part



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from cpymad.madx import Madx

# hllhc15 can be found at git@github.com:lhcopt/hllhc15.git

mad = Madx()

mad.input("""
call,file="../../../hllhc15/util/lhc.seq";
call,file="../../../hllhc15/hllhc_sequence.madx";
call,file="../../../hllhc15/toolkit/macro.madx";
seqedit,sequence=lhcb1;flatten;cycle,start=IP3;flatten;endedit;
seqedit,sequence=lhcb2;flatten;cycle,start=IP3;flatten;endedit;
exec,mk_beam(7000);
call,file="../../../hllhc15/round/opt_round_150_1500.madx";
on_x1 = 1;
on_x5 = 1;
on_disp = 0;
exec,check_ip(b1);
exec,check_ip(b2);
""")

mad.input('''

brho:=NRJ*1e9/clight;
! Octupoles, added by hand
i_oct_b1=0;
  kof.a12b1:=kmax_mo*i_oct_b1/imax_mo/brho; kof.a23b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kof.a34b1:=kmax_mo*i_oct_b1/imax_mo/brho; kof.a45b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kof.a56b1:=kmax_mo*i_oct_b1/imax_mo/brho; kof.a67b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kof.a78b1:=kmax_mo*i_oct_b1/imax_mo/brho; kof.a81b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kod.a12b1:=kmax_mo*i_oct_b1/imax_mo/brho; kod.a23b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kod.a34b1:=kmax_mo*i_oct_b1/imax_mo/brho; kod.a45b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kod.a56b1:=kmax_mo*i_oct_b1/imax_mo/brho; kod.a67b1:=kmax_mo*i_oct_b1/imax_mo/brho;
  kod.a78b1:=kmax_mo*i_oct_b1/imax_mo/brho; kod.a81b1:=kmax_mo*i_oct_b1/imax_mo/brho;
''')


mad.use(sequence="lhcb1")
mad.globals['vrf400'] = 16
mad.globals['lagrf400.b1'] = 0.5
mad.globals['lagrf400.b2'] = 0.
mad.twiss()
mad.save(sequence=['lhcb1', 'lhcb2'], beam=True, file="sequence_thick.madx")
mad.call('../../../hllhc15/toolkit/enable_crabcavities.madx')
mad.twiss()
mad.save(sequence=['lhcb1'], beam=True, file="sequence_thick_with_crabs.madx")

mad_b4 = Madx()

mad_b4.input("""
mylhcbeam = 4;
call,file="../../../hllhc15/util/lhcb4.seq";
call,file="../../../hllhc15/hllhc_sequence.madx";
call,file="../../../hllhc15/toolkit/macro.madx";
seqedit,sequence=lhcb2;flatten;cycle,start=IP3;flatten;endedit;
exec,mk_beam(7000);
call,file="../../../hllhc15/round/opt_round_150_1500.madx";
on_x1 = 1;
on_x5 = 1;
on_disp = 0;
exec,check_ip(b2);
""")
mad_b4.use(sequence="lhcb2")
mad_b4.globals['vrf400'] = 16
mad_b4.globals['lagrf400.b2'] = 0
mad_b4.twiss()
mad_b4.save(sequence=['lhcb2'], beam=True, file="sequence_thick_b4.madx")


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

import matplotlib.pyplot as plt

# Load a line and build tracker
line = xt.Line.from_json("../../test_data/hllhc15_thick/lhc_thick_with_knobs.json")
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)

# Twiss
tw = line.twiss4d()

pl = tw.plot(figlabel="tw")
pl = tw.plot(figlabel="tw")

fig = plt.figure()
ax1 = plt.subplot(211)
ax2 = plt.subplot(212, sharex=ax1)
pl1 = tw.plot(ax=ax1)
pl2 = tw.plot(yl="mux muy", ax=ax2)
fig.subplots_adjust(right=0.7)
pl1.move_legend(1.5, 1.0)
pl2.move_legend(1.5, 1.0)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

line = xt.Line.from_json(
    '../../test_data/psb_injection/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=(xt.PROTON_MASS_EV), q0=1, energy0=(2*160.+938.)*1.e6)
line.build_tracker()

df_hz = 180 # Frequency trim

tw = line.twiss()

h_rf = 1
f_rf = h_rf/tw.T_rev0

beta0 = line.particle_ref.beta0[0]

# T_rev = h_rf/f_rf
# dt = h_rf/(f_rf + df_hz) - h_rf/f_rf = h_rf/f_rf (1/(1+df_hz/f_rf) - 1)
#                                       ~= h_rf/f_rf * (1 - df_hz/f_rf -1)
#                                       = -h_rf/(f_rf^2) * df_hz
#                                       = -T_rev / f_rf * df_hz
# dzeta = -beta0 * clight * dt = circumference * df_hz / f_rf

dzeta = tw.circumference * df_hz / f_rf

line.unfreeze()
line.append_element(element=xt.ZetaShift(dzeta=dzeta), name='zeta_shift')
line.build_tracker()

tw_6d_offmom = line.twiss()

print(f'delta closed orbit: {tw_6d_offmom.delta[0]:.3e}')
# prints: delta closed orbit: 2.848e-04

# Checks
import numpy as np
eta = tw.slip_factor
f0 = 1/tw.T_rev0
delta_trim = -1/h_rf/eta/f0*df_hz

# Use 4d twiss on machine without zeta shift
line['zeta_shift'].dzeta = 0
tw_on_mom = line.twiss(delta0=0, method='4d')
tw_off_mom = line.twiss(delta0=delta_trim, method='4d')
dzeta_from_twiss = (tw_off_mom['zeta'][-1] - tw_off_mom['zeta'][0])

print(f'delta_trim: {delta_trim}')
print(f'tw_6d_offmom.delta[0]: {tw_6d_offmom.delta[0]}')
print(f'tw_6d_offmom.ptau[0]: {tw_6d_offmom.ptau[0]}')

assert np.isclose(delta_trim, tw_6d_offmom.delta[0], rtol=1e-3, atol=0)
assert np.isclose(dzeta, dzeta_from_twiss, rtol=1e-3, atol=0)



import xtrack as xt
import xdeps as xd
import numpy as np

from xtrack.mad_parser.loader import CONSTANTS

formatter = xd.refs.CompactFormatter(scope=None)
SKIP_PARAMS = ['order', 'model', '_edge_entry_model', '_edge_exit_model',
               'k0_from_h', 'h']


def _repr_arr_ref(arr_ref, formatter):
    out = []
    for ii, vv in enumerate(arr_ref._value):
        if arr_ref[ii]._expr is not None:
            out.append(f'"{arr_ref[ii]._expr._formatted(formatter)}"')
        else:
            out.append(f'{arr_ref[ii]._value:g}')

    # Trim trailing zeros
    while out and (out[-1] == '0' or out[-1] == '-0'):
        out.pop()

    return f'[{", ".join(out)}]'

def _elem_to_tokens(env, nn, formatter):

    ee = env.get(nn)
    ee_ref = env.ref[nn]

    # The fields to consider are those in the dictionary, plus knl and ksl, plus
    # anything that has an expression
    fields = list(ee.to_dict().keys())
    if hasattr(ee, 'knl'):
        fields += ['knl']
    if hasattr(ee, 'ksl'):
        fields += ['ksl']

    tt = env[nn].get_table()
    for kk in tt.name:
        if tt['expr', kk] is not None and tt['expr', kk] != 'None':
            fields.append(kk)

    params = []
    for kk in fields:
        if kk == '__class__':
            continue
        if kk in SKIP_PARAMS:
            continue
        if kk == 'knl' or kk == 'ksl':
            arr_ref = getattr(ee_ref, kk)
            vv = _repr_arr_ref(arr_ref, formatter)
            if vv != '[]':
                params.append(f'{kk}={vv}')
        elif getattr(ee_ref, kk)._expr is not None:
            params.append(f'{kk}="{getattr(ee_ref, kk)._expr._formatted(formatter)}"')
        else:
            params.append(f'{kk}={getattr(ee_ref, kk)._value:g}')

    out = {'name': nn, 'element_type': ee.__class__.__name__, 'params': params,
           'prototype': getattr(ee, 'prototype', None),
           'extra': getattr(ee, 'extra', None)}
    return out

####################
# Load the lattice #
####################

env = xt.load_madx_lattice('../../test_data/lhc_2024/lhc.seq', reverse_lines=['lhcb2'])

###################
# Handle elements #
###################

all_elems = []
for lname in env.lines.keys():
    ll = env.lines[lname]
    bb = ll.builder.flatten()
    all_elems += [cc.name for cc in bb.components]

elem_tokens = {}
for nn in all_elems:
    elem_tokens[nn] = _elem_to_tokens(env, nn, formatter)

while True:
    added = False
    for nn in list(elem_tokens.keys()):
        if elem_tokens[nn]['prototype'] is not None:
            parent_name = elem_tokens[nn]['prototype']
            if parent_name not in elem_tokens:
                elem_tokens[parent_name] = _elem_to_tokens(env, parent_name, formatter)
                added = True
    if not added:
        break

# Some customizations
elem_tokens['multipole']['params'].append('knl=[0,0,0,0,0,0]')

# remove length when length_straight is present
for nn in elem_tokens:
    if 'length_straight' in elem_tokens[nn]['params']:
        elem_tokens[nn]['params'] = [pp for pp in elem_tokens[nn]['params'] if pp != 'length']

# populate diff params
for nn in elem_tokens:
    diff_params = []
    if elem_tokens[nn]['prototype'] is not None:
        parent_name = elem_tokens[nn]['prototype']
        parent_params = elem_tokens[parent_name]['params']
        elem_params = set(elem_tokens[nn]['params'])
        for pp in elem_params:
            if pp not in parent_params:
                diff_params.append(pp)
    else:
        diff_params = elem_tokens[nn]['params']
    elem_tokens[nn]['diff_params'] = diff_params

# Build def instruction
for nn in elem_tokens:
    out_parts = []
    out_parts.append(f'env.new("{nn}"')
    if elem_tokens[nn]['prototype'] is not None:
        out_parts.append(f'"{elem_tokens[nn]["prototype"]}"')
    else:
        out_parts.append(f'"{elem_tokens[nn]["element_type"]}"')
    if len(elem_tokens[nn]['diff_params']) > 0:
        out_parts += elem_tokens[nn]['diff_params']
    if elem_tokens[nn]['extra'] is not None:
        out_parts.append(f'extra={elem_tokens[nn]["extra"]}')
    elem_tokens[nn]['def_instruction'] = ', '.join(out_parts) + ')'

# Sort based on hierarchy
sorted_elems = []
def _add_elem(nn):
    if elem_tokens[nn]['prototype'] is not None:
        _add_elem(elem_tokens[nn]['prototype'])
    if nn not in sorted_elems:
        sorted_elems.append(nn)
for nn in elem_tokens:
    _add_elem(nn)

tt_edefs = xt.Table({
    'name': np.array(sorted_elems),
    'element_type': np.array([elem_tokens[nn]['element_type'] for nn in sorted_elems]),
    'prototype': np.array([elem_tokens[nn]['prototype'] for nn in sorted_elems]),
    'def_instruction': np.array([elem_tokens[nn]['def_instruction'] for nn in sorted_elems]),
})

tt_gen0 = tt_edefs.rows[tt_edefs['prototype'] == None]
tt_gen0.gen_name = 'Xsuite types'

generation_tree = [{'Xsuite types': tt_gen0}]
while True:
    print(f'Generation {len(generation_tree)}')
    last_gen = generation_tree[-1]
    names_last_gen = []
    for nn in last_gen.keys():
        names_last_gen += list(last_gen[nn]['name'])
    this_gen = {}
    added = False
    for nn in names_last_gen:
        tt_gen = tt_edefs.rows[tt_edefs['prototype'] == nn]
        if len(tt_gen) > 0:
            tt_gen.gen_name = nn
            this_gen[nn] = tt_gen
            added = True
    if added:
        generation_tree.append(this_gen)
    else:
        break

# Flatten generations
generations = []
for gg in generation_tree:
    for nn in gg.keys():
        generations.append(gg[nn])

# Build elem def part (with generations)
elem_def_lines = []
for tt_gen in generations:
    elem_def_lines.append(f'\n# Elements of type: {tt_gen.gen_name}')
    for nn in tt_gen.name:
        elem_def_lines.append(elem_tokens[nn]['def_instruction'])


# # Build elem def part (no generations)
# elem_def_lines = []
# for nn in sorted_elems:
#     elem_def_lines.append(elem_tokens[nn]['def_instruction'])

elem_def_part = '\n'.join(elem_def_lines)

###################
# Handle builders #
###################

builder_lines = []
for lname in env.lines.keys():
    builder_lines.append(f'# Builder for line: {lname}')
    bb = env.lines[lname].builder.flatten()
    builder_lines.append(f'{lname} = env.new_builder(name="{lname}")')
    for cc in bb.components:
        cc_tokens=[]
        for kk, vv in cc.__dict__.items():
            if vv is None or kk == 'name':
                continue
            if hasattr(vv, '_expr'): # is expression
                # Get string representation of expression
                env['__temp__'] = vv
                vv = env.ref["__temp__"]._expr._formatted(formatter)
            cc_tokens.append(f'{kk}="{vv}"')
        builder_lines.append(f'{lname}.place("{cc.name}", ' + ', '.join(cc_tokens) + ')')
    builder_lines.append(f'{lname}.build()')
    builder_lines.append('')

builder_part = '\n'.join(builder_lines)

####################
# Handle variables #
####################

ttvars = env.vars.get_table()
const = set(CONSTANTS.keys())
lattice_parameters = [nn for nn in ttvars.name if nn not in const]

tt_lattice_pars_all = ttvars.rows[lattice_parameters]
mask_keep = (tt_lattice_pars_all['expr'] != None) | (tt_lattice_pars_all['value'] != 0)
tt_lattice_pars = tt_lattice_pars_all.rows[mask_keep]

lines_vars = []
for nn in tt_lattice_pars.name:
    if nn == '__temp__':
        continue
    if tt_lattice_pars['expr', nn] is not None:
        lines_vars.append(f'env["{nn}"] = "{tt_lattice_pars["expr", nn]}"')
    else:
        lines_vars.append(f'env["{nn}"] = {tt_lattice_pars["value", nn]}')
lines_vars.append('')
vars_part = '\n'.join(lines_vars)

#####################
# Assemble the file #
#####################

preamble = '''import xtrack as xt
env = xt.get_environment()
env.vars.default_to_zero=True
'''

postamble = '''

env.vars.default_to_zero=False

'''

file_content = '\n'.join([
    preamble,
   '#############',
   '# Variables #',
   '#############',
   '',
   vars_part,
   '############',
   '# Elements #',
   '############',
   elem_def_part,
   '',
   '##############',
   '# Beam lines #',
   '##############',
   '',
   builder_part,
   postamble])

with open('lhc_seq.py', 'w') as ff:
    ff.write(file_content)